from __future__ import annotations

from abc import ABC, abstractmethod
from ast import (
    AST,
    Break,
    Continue,
    ExceptHandler,
    For,
    If,
    NodeVisitor,
    Try,
    While,
    With,
    parse,
    withitem,
)
from typing import Any, List, Tuple

from pygit2 import Blob, Oid, Tree
from radon.complexity import cc_visit
from radon.raw import Module, analyze

from .visitableobject import VisitableBlob, VisitableTree


class NestingASTVisitor(NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.n: int = 0
        self.max_n: int = self.n

    def visit(self, node: AST) -> NestingASTVisitor:
        self.max_n = max(self.max_n, self.n)
        super().visit(node)
        return self

    def nesting_visit(self, node: AST) -> None:
        self.n += 1
        super().generic_visit(node)
        self.n -= 1

    def visit_If(self, node: If) -> None:
        return self.nesting_visit(node)

    def visit_For(self, node: For) -> None:
        return self.nesting_visit(node)

    def visit_While(self, node: While) -> None:
        return self.nesting_visit(node)

    def visit_Break(self, node: Break) -> None:
        return self.nesting_visit(node)

    def visit_Continue(self, node: Continue) -> None:
        return self.nesting_visit(node)

    def visit_Try(self, node: Try) -> None:
        return self.nesting_visit(node)

    def visit_ExceptHandler(self, node: ExceptHandler) -> None:
        return self.nesting_visit(node)

    def visit_With(self, node: With) -> None:
        return self.nesting_visit(node)

    def visit_withitem(self, node: withitem) -> None:
        return self.nesting_visit(node)

    @property
    def result(self) -> int:
        return self.max_n


class TreeVisitor(ABC):
    @abstractmethod
    def visitBlob(self, blob: VisitableBlob) -> TreeVisitor:
        pass

    def visitTree(self, tree: VisitableTree) -> TreeVisitor:
        for obj in tree.obj:
            if isinstance(obj, Tree):
                VisitableTree(obj).accept(self)
            elif isinstance(obj, Blob):
                VisitableBlob(obj).accept(self)
        return self

    @property
    @abstractmethod
    def result(self) -> Any:
        pass


class LocVisitor(TreeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.n: int = 0

    def visitBlob(self, blob: VisitableBlob) -> LocVisitor:
        if not blob.obj.is_binary:
            self.n += len(blob.obj.data.split(b"\n"))
        return self

    @property
    def result(self) -> int:
        return self.n


class FilecountVisitor(TreeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.n: int = 0

    def visitBlob(self, blob: VisitableBlob) -> FilecountVisitor:
        self.n += 1
        return self

    @property
    def result(self) -> int:
        return self.n


class ComplexityVisitor(TreeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.complexities: List[Tuple[Oid, str]] = []

    def visitBlob(self, blob: VisitableBlob) -> ComplexityVisitor:
        if blob.obj.name.endswith(".py"):
            if complexities := cc_visit(blob.obj.data):
                self.complexities.append(
                    (blob.obj.id, max(obj.complexity for obj in complexities))
                )
        return self

    @property
    def result(self) -> List[Tuple[Oid, str]]:
        return self.complexities


class RawVisitor(TreeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.metrics: List[Tuple[Oid, Module]] = []

    def visitBlob(self, blob: VisitableBlob) -> RawVisitor:
        if blob.obj.name.endswith(".py"):
            data: Module = analyze(blob.obj.data.decode())
            self.metrics.append((blob.obj.id, data))
        return self

    @property
    def result(self) -> List[Tuple[Oid, Module]]:
        return self.metrics


class NestingVisitor(TreeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.metrics: List[Tuple[Oid, int]] = []

    def visitBlob(self, blob: VisitableBlob) -> NestingVisitor:
        if blob.obj.name.endswith(".py"):
            m = parse(blob.obj.data)
            self.metrics.append((blob.obj.id, NestingASTVisitor().visit(m).result))
        return self

    @property
    def result(self) -> List[Tuple[Oid, int]]:
        return self.metrics



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterable

from pygit2 import (
    GIT_SORT_REVERSE,
    GIT_SORT_TOPOLOGICAL,
    Commit,
    Repository,
    Walker,
    clone_repository,
)
from typer import Typer, echo

from .treevisitor import (
    ComplexityVisitor,
    FilecountVisitor,
    LocVisitor,
    NestingVisitor,
    RawVisitor,
)
from .visitableobject import VisitableTree

app = Typer()


def walk_commits(
    repo: Repository, branch_name: str, simplify_first_parent: bool
) -> Iterable[Commit]:
    branch = repo.branches[branch_name]
    walker: Walker = repo.walk(
        branch.peel().id, GIT_SORT_TOPOLOGICAL | GIT_SORT_REVERSE
    )
    if simplify_first_parent:
        walker.simplify_first_parent()
    yield from walker


@app.command()
def clone(
    url: str,
    path: Path,
) -> None:
    "Clone a repository to a path."
    clone_repository(url, path, bare=True)
    echo(f"Cloned {url} to {path}")


@app.command()
def branch(path: Path, local: bool = True, remote: bool = True) -> None:
    repo = Repository(path)
    branches: Iterable[str] = tuple()
    if local and remote:
        branches = repo.branches
    elif local and not remote:
        branches = repo.branches.local
    elif not local and remote:
        branches = repo.branches.remote
    for branch_name in branches:
        echo(branch_name)


@app.command()
def complexity(path: Path, simplify_first_parent: bool = True) -> None:
    """Get the maximum complexity per file per commit per branch.

    Currently, only Python files are supported."""
    repo = Repository(path)
    for branch_name in repo.branches:
        for commit in walk_commits(repo, branch_name, simplify_first_parent):
            for blob_id, result in (
                ComplexityVisitor().visitTree(VisitableTree(commit.tree)).result
            ):
                echo(
                    "{:s},{:s},{:s},{:d}".format(
                        branch_name, str(commit.id), str(blob_id), result
                    )
                )


@app.command()
def raw(path: Path, simplify_first_parent: bool = True) -> None:
    """Get raw metrics per file per commit per branch.

    Currently, only Python files are supported.
    Raw metrics are:
    1. LOC:
       The number of lines of code (total)
    2. LLOC:
       The number of logical lines of code
    3. SLOC:
       The number of source lines of code (not necessarily corresponding to the LLOC)
    """
    repo = Repository(path)
    for branch_name in repo.branches:
        for commit in walk_commits(repo, branch_name, simplify_first_parent):
            for blob_id, result in (
                RawVisitor().visitTree(VisitableTree(commit.tree)).result
            ):
                echo(
                    "{:s},{:s},{:s},{:d},{:d},{:d}".format(
                        branch_name,
                        str(commit.id),
                        str(blob_id),
                        result.loc,
                        result.lloc,
                        result.sloc,
                    )
                )


@app.command()
def nesting(path: Path, simplify_first_parent: bool = True) -> None:
    """Get nesting level per file per commit per branch."""
    repo = Repository(path)
    for branch_name in repo.branches:
        for commit in walk_commits(repo, branch_name, simplify_first_parent):
            for blob_id, result in (
                NestingVisitor().visitTree(VisitableTree(commit.tree)).result
            ):
                echo(
                    "{:s},{:s},{:s},{:d}".format(
                        branch_name,
                        str(commit.id),
                        str(blob_id),
                        result,
                    )
                )


@app.command()
def filecount(path: Path, simplify_first_parent: bool = True) -> None:
    """Get the number of files per commit in each branch in a local repository."""
    repo = Repository(path)
    for branch_name in repo.branches:
        for commit in walk_commits(repo, branch_name, simplify_first_parent):
            echo(
                "{:s},{:s},{:d}".format(
                    branch_name,
                    str(commit.id),
                    FilecountVisitor().visitTree(VisitableTree(commit.tree)).result,
                )
            )


@app.command()
def remote_filecount(
    url: str, checkout_branch: str, simplify_first_parent: bool = True
) -> None:
    """Get the number of files per commit."""
    with TemporaryDirectory() as tmpdirname:
        repo: Repository = clone_repository(
            url, tmpdirname, checkout_branch=checkout_branch, bare=True
        )
        for commit in walk_commits(repo, repo.head.shorthand, simplify_first_parent):
            echo(
                "{:s},{:s},{:d}".format(
                    repo.head.shorthand,
                    str(commit.id),
                    FilecountVisitor().visitTree(VisitableTree(commit.tree)).result,
                )
            )


@app.command()
def remote_loc(
    url: str, checkout_branch: str, simplify_first_parent: bool = True
) -> None:
    """Get the lines of code per commit."""
    with TemporaryDirectory() as tmpdirname:
        repo: Repository = clone_repository(
            url, tmpdirname, checkout_branch=checkout_branch, bare=True
        )
        for commit in walk_commits(repo, repo.head.shorthand, simplify_first_parent):
            echo(
                "{:s},{},{:d}".format(
                    repo.head.shorthand,
                    str(commit.id),
                    LocVisitor().visitTree(VisitableTree(commit.tree)).result,
                )
            )
